package com.dbxml.db.common.adapters;

/*
 * dbXML - Native XML Database
 * Copyright (C) 1999-2004  The dbXML Group, L.L.C.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: SAXAdapter.java,v 1.4 2004/02/12 00:17:56 bradford Exp $
 */

import com.dbxml.db.core.Collection;
import com.dbxml.db.core.DBException;
import com.dbxml.db.core.FaultCodes;
import com.dbxml.db.core.adapter.SimpleAdapter;
import com.dbxml.db.core.data.Key;
import com.dbxml.db.core.transaction.Transaction;
import com.dbxml.xml.SymbolTable;
import com.dbxml.xml.dtsm.DTSMException;
import com.dbxml.xml.dtsm.DTSMHelper;
import com.dbxml.xml.dtsm.DocumentTable;
import com.dbxml.xml.dtsm.TableBuilder;
import com.dbxml.xml.sax.DTSMContentHandler;
import org.xml.sax.ContentHandler;

/**
 * SAXAdapter adapts a dbXML Collection to support the management of
 * Documents using the SAX API.
 */

public final class SAXAdapter extends SimpleAdapter {
   public SAXAdapter(Collection col) {
      super(col);
   }

   private ContentHandler putDocument(Transaction tx, Object key) throws DBException {
      SymbolTable symbols = col.getSymbols();
      TableBuilder builder = new TableBuilder(symbols);
      return new AdapterContentHandler(tx, key, builder);
   }

   /**
    * setDocument sets a Document using SAX events that are fed to the
    * returned ContentHandler.  The document is stored in the Collection
    * using the provided Key.
    *
    * @param tx The controlling Transaction
    * @param key The Key to use
    * @return a ContentHandler for feeding SAX events
    * @throws DBException If a Database Exception occurs
    */
   public ContentHandler setDocument(Transaction tx, Object key) throws DBException {
      return putDocument(tx, key);
   }

   /**
    * setDocument sets a Document using SAX events that are fed to the
    * returned ContentHandler.  The document is stored in the Collection
    * using a newly generated Key.
    *
    * @param tx The controlling Transaction
    * @return a ContentHandler for feeding SAX events
    * @throws DBException If a Database Exception occurs
    */
   public ContentHandler insertDocument(Transaction tx) throws DBException {
      Key key = col.createNewOID();
      return putDocument(tx, key);
   }

   /**
    * insertDocument sets a Document using SAX events that are fed to the
    * returned ContentHandler.  The document is stored in the Collection
    * using the provided Key.
    *
    * @param tx The controlling Transaction
    * @param key The Key to use
    * @return a ContentHandler for feeding SAX events
    * @throws DBException If a Database Exception occurs
    */
   public ContentHandler insertDocument(Transaction tx, Object key) throws DBException {
      return putDocument(tx, key);
   }

   /**
    * getDocument instantiates and returns a DOM Document based on the
    * provided Key.
    *
    * @param tx The controlling Transaction
    * @param key The Document Key
    * @return a DOM Document
    * @throws DBException If a Database Exception occurs
    */
   public void getDocument(Transaction tx, Object key, ContentHandler handler) throws DBException {
      try {
         DocumentTable dt = col.getDocument(tx, key);
         if ( dt != null )
            DTSMHelper.tableToSAX(dt, handler);
      }
      catch ( DTSMException e ) {
         throw new DBException(FaultCodes.COL_CANNOT_RETRIEVE, "Error retrieving object '" + key + "'", e);
      }
   }


   /**
    * AdapterContentHandler
    */

   private class AdapterContentHandler extends DTSMContentHandler {
      public Transaction tx;
      public Object key;

      public AdapterContentHandler(Transaction tx, Object key, TableBuilder builder) {
         super(builder);
         this.tx = tx;
         this.key = key;
      }

      public void complete() {
         try {
            DocumentTable dt = getTableBuilder().buildDocumentTable();
            col.setDocument(tx, key, dt);
         }
         catch ( DBException e ) {
            /** @todo Uh-oh... What do we do? */
            e.printStackTrace(System.err);
         }
      }
   }
}
